from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.date_time_label_formats import DateTimeLabelFormats

from highcharts_core.options.axes.generic import GenericAxis
from highcharts_core.options.axes.breaks import AxisBreak
from highcharts_core.options.axes.plot_bands import PlotBand, PlotLine
from highcharts_core.options.axes.title import AxisTitle


class NumericAxis(GenericAxis):
    """Base class that is used for defining numeric axes."""

    def __init__(self, **kwargs):
        self._align_ticks = None
        self._allow_decimals = None
        self._alternate_grid_color = None
        self._breaks = None
        self._categories = None
        self._date_time_label_formats = None
        self._linked_to = None
        self._min_range = None
        self._min_tick_interval = None
        self._offset = None
        self._opposite = None
        self._pane = None
        self._plot_bands = None
        self._plot_lines = None
        self._reversed_stacks = None
        self._title = None
        self._zoom_enabled = None

        self.align_ticks = kwargs.get('align_ticks', None)
        self.allow_decimals = kwargs.get('allow_decimals', None)
        self.alternate_grid_color = kwargs.get('alternate_grid_color', None)
        self.breaks = kwargs.get('breaks', None)
        self.categories = kwargs.get('categories', None)
        self.date_time_label_formats = kwargs.get('date_time_label_formats', None)
        self.linked_to = kwargs.get('linked_to', None)
        self.min_range = kwargs.get('min_range', None)
        self.min_tick_interval = kwargs.get('min_tick_interval', None)
        self.offset = kwargs.get('offset', None)
        self.opposite = kwargs.get('opposite', None)
        self.pane = kwargs.get('pane', None)
        self.plot_bands = kwargs.get('plot_bands', None)
        self.plot_lines = kwargs.get('plot_lines', None)
        self.reversed_stacks = kwargs.get('reversed_stacks', None)
        self.title = kwargs.get('title', None)
        self.zoom_enabled = kwargs.get('zoom_enabled', None)

        super().__init__(**kwargs)

    @property
    def align_ticks(self) -> Optional[bool]:
        """If ``True`` and using multiple axes, the ticks of two or more opposite axes
        will automatically be aligned by adding ticks to the axis or axes with the least
        ticks, as if ``tick_amount`` were specified. This can be prevented by setting
        ``align_ticks`` to ``False``.

        Defaults to ``True``.

        .. hint::

          If the grid lines look messy, it's a good idea to hide them for the secondary
          axis by setting ``grid_line_width`` to ``0``.

        .. warning::

          If ``start_on_tick`` or ``end_on_tick`` in the axis options are set to
          ``False``, then tick alignment will be disabled for the axis.

        .. warning::

          Aways disabled for logarithmic axes.

        :returns: Flag indicating whether ot align ticks along the axes.
        :rtype: :class:`bool <python:bool>`  or :obj:`None <python:None>`
        """
        return self._align_ticks

    @align_ticks.setter
    def align_ticks(self, value):
        if value is None:
            self._align_ticks = None
        else:
            self._align_ticks = bool(value)

    @property
    def allow_decimals(self) -> Optional[bool]:
        """If ``True``, allow decimals in the axis' ticks. If :obj:`None <python:None>`,
        decimals are allowed on small scale axes. Defaults to :obj:`None <python:None>`.

        .. hint::

          When counting integers, like persons or hits on a web page, decimals should be
          avoided in the labels.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_decimals

    @allow_decimals.setter
    def allow_decimals(self, value):
        if value is None:
            self._allow_decimals = None
        else:
            self._allow_decimals = bool(value)

    @property
    def alternate_grid_color(self) -> Optional[str | Gradient | Pattern]:
        """When using an alternate grid color, a band is painted across the plot area
        between every other grid line. This setting determines the color of that band.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._alternate_grid_color

    @alternate_grid_color.setter
    def alternate_grid_color(self, value):
        from highcharts_core import utility_functions
        self._alternate_grid_color = utility_functions.validate_color(value)

    @property
    def breaks(self) -> Optional[List[AxisBreak]]:
        """A collection that defines breaks to render in the axis. Defaults to
        :obj:`None <python:None>`.

        .. note::

          The sections defined as breaks will be left out and all the points shifted
          closer to each other.

        :rtype: :class:`list <python:list>` of :class:`AxisBreak`, or
          :obj:`None <python:None>`
        """
        return self._breaks

    @breaks.setter
    @class_sensitive(AxisBreak, force_iterable = True)
    def breaks(self, value):
        self._breaks = value

    @property
    def categories(self) -> Optional[List[str]]:
        """If categories are present for the x-axis, displays category names along the
        x-axis instead of numerical values. Defaults to :obj:`None <python:None>`.

        .. hint::

          While categories can also be defined by giving each data point a name and
          setting axis type to ``'category'``, if you have multiple series, best practice
          remains using the :meth:`categories <NumericAxis.categories>` setting for ease
          of maintenance.

        :rtype: :class:`list <python:list>` of :class:`str <python:str>`, or
          :obj:`None <python:None>`
        """
        return self._categories

    @categories.setter
    def categories(self, value):
        if not value:
            self._categories = None
        else:
            self._categories = [validators.string(x, allow_empty = True) or '' for x in validators.iterable(value)]

    @property
    def date_time_label_formats(self) -> Optional[DateTimeLabelFormats]:
        """For a datetime axis, the scale will automatically adjust to the appropriate
        unit. This setting gives the default string representations used for each unit.
        Defaults to :obj:`None <python:None>`.

        .. note::

          For intermediate values, different units may be used, for example the day unit
          can be used on midnight and hour unit  used for intermediate values on the same
          axis.

        .. seealso::

          * :class:`DateTimeLabelFormats`

        :rtype: :class:`DateTimeLabelFormats` or :obj:`None <python:None>`
        """
        return self._date_time_label_formats

    @date_time_label_formats.setter
    @class_sensitive(DateTimeLabelFormats)
    def date_time_label_formats(self, value):
        self._date_time_label_formats = value

    @property
    def linked_to(self) -> Optional[int]:
        """The index of another axis that this axis is linked to.

        When an axis is linked to a master axis, it will take the same extremes as the
        master as assigned by :meth:`NumericAxis.min` or :meth:`NumericAxis.max` or by the
        (JavaScript) ``setExtremes()`` method. It can be used to show additional info, or
        to ease reading the chart by duplicating the scales.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._linked_to

    @linked_to.setter
    def linked_to(self, value):
        self._linked_to = validators.integer(value,
                                             allow_empty = True,
                                             minimum = 0)

    @property
    def min_range(self) -> Optional[int | float | Decimal]:
        """The minimum range to display on this axis. The entire axis will not be allowed
        to span over a smaller interval than this, particularly when zooming. Defaults to
        :obj:`None <python:None>`.

        For example, for a datetime axis the main unit is milliseconds. If ``min_range``
        is set to ``3600000``, the user won't be able to zoom in more than to one hour.

        If :obj:`None <pythoN:None>`, the ``min_range`` for the x-axis will automatically
        be set to five times the smallest interval between any of the data points.

        On a logarithmic axis, the unit for the minimum range is the power. So a
        ``min_range`` of ``1`` means that the axis can be zoomed to 10-100, 100-1000,
        1000-10,000 etc.

        .. warning::

          The following settings also impact how the extremes of the axis are computed:

            * :meth:`NumericAxis.min_padding`
            * :meth:`NumericAxis.max_padding`
            * :meth:`NumericAxis.start_on_tick`
            * :meth:`NumericAxis.end_on_tick`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_range

    @min_range.setter
    def min_range(self, value):
        self._min_range = validators.numeric(value, allow_empty = True)

    @property
    def min_tick_interval(self) -> Optional[int | float | Decimal]:
        """The minimum tick interval allowed in axis values. Defaults to
        :obj:`None <python:None>`, which applies the closest distance between two points
        on the axis.

        For example, this setting can be used to prevent the axis from showing hours as
        minor tick marks.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_tick_interval

    @min_tick_interval.setter
    def min_tick_interval(self, value):
        self._min_tick_interval = validators.numeric(value, allow_empty = True)

    @property
    def offset(self) -> Optional[int | float | Decimal]:
        """The distance in pixels from the plot area to the axis line. Defaults to
        :obj:`None <python:None>`.

        A positive offset moves the axis (along with its line, labels, and ticks) away
        from the plot area. This is typically used when two or more axes are displayed on
        the same side of the plot.

        With multiple axes, the ``offset`` is dynamically adjusted to avoid collision,
        though this can be overridden by setting ``offset`` explicitly.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = validators.numeric(value, allow_empty = True)

    @property
    def opposite(self) -> Optional[bool]:
        """If ``True``, displays the axis on the opposite side of where it would normally
        appear. Defaults to ``False``.

        Vertical axes would normally appear on the left side of the chart, while
        horizontal axes would normally appear on the bottom of the chart. Thus, the
        opposite side would be the right and top respectively.

        .. hint::

          This feature is typically used with dual or multiple axes.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._opposite

    @opposite.setter
    def opposite(self, value):
        if value is None:
            self._opposite = None
        else:
            self._opposite = bool(value)

    @property
    def pane(self) -> Optional[int]:
        """Indicates the index in the panes array. When :obj:`None <python:None>`, the
        first pane will be used. Defaults to :obj:`None <python:None>`.

        .. warning::

          Used for circular gauges and polar charts.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._pane

    @pane.setter
    def pane(self, value):
        self._pane = validators.integer(value,
                                        allow_empty = True,
                                        minimum = 0)

    @property
    def plot_bands(self) -> Optional[List[PlotBand]]:
        """A collection of colored bands stretching across the plot area which can be used
        to visually mark an interval along the axis. Defaults to
        :obj:`None <python:None>`.

        .. note::

          In a :term:`gauge chart`, a plot band on the :class:`YAxis` (value axis) will
          stretch along the perimeter of the gauge.

        :rtype: :class:`list <python:list>` of :class:`PlotBand` instances, or
          :obj:`None <python:None>`
        """
        return self._plot_bands

    @plot_bands.setter
    @class_sensitive(PlotBand, force_iterable = True)
    def plot_bands(self, value):
        self._plot_bands = value

    @property
    def plot_lines(self) -> Optional[List[PlotLine]]:
        """A collection of lines stretching across the plot area which can be used
        to visually mark specific values on the axis. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`list <python:list>` of :class:`PlotLine` instances, or
          :obj:`None <python:None>`
        """
        return self._plot_lines

    @plot_lines.setter
    @class_sensitive(PlotLine, force_iterable = True)
    def plot_lines(self, value):
        self._plot_lines = value

    @property
    def reversed_stacks(self) -> Optional[bool]:
        """This option determines how stacks should be ordered within a group. Setting
        this to ``True`` will reverse the order of series within a group relative to
        its automatic ordering.

        For example, if :class:`XAxis.reversed`` is ``True``, then the default behavior is
        to reverse the order of the series so that the first series comes last in a group.
        To retain the original order, set this option to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reversed_stacks

    @reversed_stacks.setter
    def reversed_stacks(self, value):
        if value is None:
            self._reversed_stacks = None
        else:
            self._reversed_stacks = bool(value)

    @property
    def title(self) -> Optional[AxisTitle]:
        """The axis title, displayed next to the axis line. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`AxisTitle` or :obj:`None <python:None>`
        """
        return self._title

    @title.setter
    @class_sensitive(AxisTitle)
    def title(self, value):
        self._title = value

    @property
    def zoom_enabled(self) -> Optional[bool]:
        """If ``True``, the axis supports zooming in to view more granular sub-sections
        of the data. If ``False``, the axis will not zoom. Defaults to ``True``.

        .. hint::

          If :meth:`Chart.zoom_type` is set, setting this value to ``False`` can override
          the ability to zoom in on this specific axis.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._zoom_enabled

    @zoom_enabled.setter
    def zoom_enabled(self, value):
        if value is None:
            self._zoom_enabled = None
        else:
            self._zoom_enabled = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'align_ticks': as_dict.get('alignTicks', None),
            'allow_decimals': as_dict.get('allowDecimals', None),
            'alternate_grid_color': as_dict.get('alternateGridColor', None),
            'angle': as_dict.get('angle', None),
            'breaks': as_dict.get('breaks', None),
            'categories': as_dict.get('categories', None),
            'ceiling': as_dict.get('ceiling', None),
            'class_name': as_dict.get('className', None),
            'crossing': as_dict.get('crossing', None),
            'date_time_label_formats': as_dict.get('dateTimeLabelFormats', None),
            'end_on_tick': as_dict.get('endOnTick', None),
            'events': as_dict.get('events', None),
            'floor': as_dict.get('floor', None),
            'grid_line_color': as_dict.get('gridLineColor', None),
            'grid_line_dash_style': as_dict.get('gridLineDashStyle', None),
            'grid_line_interpolation': as_dict.get('gridLineInterpolation', None),
            'grid_line_width': as_dict.get('gridLineWidth', None),
            'grid_z_index': as_dict.get('gridZIndex', None),
            'id': as_dict.get('id', None),
            'labels': as_dict.get('labels', None),
            'linked_to': as_dict.get('linkedTo', None),
            'margin': as_dict.get('margin', None),
            'max': as_dict.get('max', None),
            'max_padding': as_dict.get('maxPadding', None),
            'min': as_dict.get('min', None),
            'minor_grid_line_color': as_dict.get('minorGridLineColor', None),
            'minor_grid_line_dash_style': as_dict.get('minorGridLineDashStyle', None),
            'minor_grid_line_width': as_dict.get('minorGridLineWidth', None),
            'minor_tick_color': as_dict.get('minorTickColor', None),
            'minor_tick_interval': as_dict.get('minorTickInterval', None),
            'minor_tick_length': as_dict.get('minorTickLength', None),
            'minor_tick_position': as_dict.get('minorTickPosition', None),
            'minor_ticks': as_dict.get('minorTicks', None),
            'minor_tick_width': as_dict.get('minorTickWidth', None),
            'min_padding': as_dict.get('minPadding', None),
            'min_range': as_dict.get('minRange', None),
            'min_tick_interval': as_dict.get('minTickInterval', None),
            'offset': as_dict.get('offset', None),
            'opposite': as_dict.get('opposite', None),
            'pane': as_dict.get('pane', None),
            'panning_enabled': as_dict.get('panningEnabled', None),
            'plot_bands': as_dict.get('plotBands', None),
            'plot_lines': as_dict.get('plotLines', None),
            'reversed': as_dict.get('reversed', None),
            'reversed_stacks': as_dict.get('reversedStacks', None),
            'show_first_label': as_dict.get('showFirstLabel', None),
            'show_last_label': as_dict.get('showLastLabel', None),
            'soft_max': as_dict.get('softMax', None),
            'soft_min': as_dict.get('softMin', None),
            'start_of_week': as_dict.get('startOfWeek', None),
            'start_on_tick': as_dict.get('startOnTick', None),
            'tick_amount': as_dict.get('tickAmount', None),
            'tick_color': as_dict.get('tickColor', None),
            'tick_interval': as_dict.get('tickInterval', None),
            'tick_length': as_dict.get('tickLength', None),
            'tickmark_placement': as_dict.get('tickmarkPlacement', None),
            'tick_pixel_interval': as_dict.get('tickPixelInterval', None),
            'tick_position': as_dict.get('tickPosition', None),
            'tick_positioner': as_dict.get('tickPositioner', None),
            'tick_positions': as_dict.get('tickPositions', None),
            'tick_width': as_dict.get('tickWidth', None),
            'title': as_dict.get('title', None),
            'type': as_dict.get('type', None),
            'unique_names': as_dict.get('uniqueNames', None),
            'units': as_dict.get('units', None),
            'visible': as_dict.get('visible', None),
            'z_index': as_dict.get('zIndex', None),
            'zoom_enabled': as_dict.get('zoomEnabled', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'alignTicks': self.align_ticks,
            'allowDecimals': self.allow_decimals,
            'alternateGridColor': self.alternate_grid_color,
            'angle': self.angle,
            'breaks': self.breaks,
            'categories': self.categories,
            'ceiling': self.ceiling,
            'className': self.class_name,
            'crossing': self.crossing,
            'dateTimeLabelFormats': self.date_time_label_formats,
            'endOnTick': self.end_on_tick,
            'events': self.events,
            'floor': self.floor,
            'gridLineColor': self.grid_line_color,
            'gridLineDashStyle': self.grid_line_dash_style,
            'gridLineInterpolation': self.grid_line_interpolation,
            'gridLineWidth': self.grid_line_width,
            'gridZIndex': self.grid_z_index,
            'id': self.id,
            'labels': self.labels,
            'linkedTo': self.linked_to,
            'margin': self.margin,
            'max': self.max,
            'maxPadding': self.max_padding,
            'min': self.min,
            'minorGridLineColor': self.minor_grid_line_color,
            'minorGridLineDashStyle': self.minor_grid_line_dash_style,
            'minorGridLineWidth': self.minor_grid_line_width,
            'minorTickColor': self.minor_tick_color,
            'minorTickInterval': self.minor_tick_interval,
            'minorTickLength': self.minor_tick_length,
            'minorTickPosition': self.minor_tick_position,
            'minorTicks': self.minor_ticks,
            'minorTickWidth': self.minor_tick_width,
            'minPadding': self.min_padding,
            'minRange': self.min_range,
            'minTickInterval': self.min_tick_interval,
            'offset': self.offset,
            'opposite': self.opposite,
            'pane': self.pane,
            'panningEnabled': self.panning_enabled,
            'plotBands': self.plot_bands,
            'plotLines': self.plot_lines,
            'reversed': self.reversed,
            'reversedStacks': self.reversed_stacks,
            'showFirstLabel': self.show_first_label,
            'showLastLabel': self.show_last_label,
            'softMax': self.soft_max,
            'softMin': self.soft_min,
            'startOfWeek': self.start_of_week,
            'startOnTick': self.start_on_tick,
            'tickAmount': self.tick_amount,
            'tickColor': self.tick_color,
            'tickInterval': self.tick_interval,
            'tickLength': self.tick_length,
            'tickmarkPlacement': self.tickmark_placement,
            'tickPixelInterval': self.tick_pixel_interval,
            'tickPosition': self.tick_position,
            'tickPositioner': self.tick_positioner,
            'tickPositions': self.tick_positions,
            'tickWidth': self.tick_width,
            'title': self.title,
            'type': self.type,
            'uniqueNames': self.unique_names,
            'units': self.units,
            'visible': self.visible,
            'zIndex': self.z_index,
            'zoomEnabled': self.zoom_enabled,
        }

        return untrimmed



from typing import Optional, List
from decimal import Decimal
import datetime

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.events import AxisEvents
from highcharts_core.utility_classes.javascript_functions import CallbackFunction

from highcharts_core.options.axes.accessibility import AxisAccessibility
from highcharts_core.options.axes.labels import AxisLabelOptions

NoneType = type(None)


class GenericAxis(HighchartsMeta):
    """Base class that is used for defining axis classes."""

    def __init__(self, **kwargs):
        self._accessibility = None
        self._angle = None
        self._ceiling = None
        self._class_name = None
        self._crossing = None
        self._end_on_tick = None
        self._events = None
        self._floor = None
        self._grid_line_color = None
        self._grid_line_dash_style = None
        self._grid_line_interpolation = None
        self._grid_line_width = None
        self._grid_z_index = None
        self._id = None
        self._labels = None
        self._margin = None
        self._max = None
        self._max_padding = None
        self._min = None
        self._minor_grid_line_color = None
        self._minor_grid_line_dash_style = None
        self._minor_grid_line_width = None
        self._minor_tick_color = None
        self._minor_tick_interval = None
        self._minor_tick_length = None
        self._minor_tick_position = None
        self._minor_ticks = None
        self._minor_tick_width = None
        self._min_padding = None
        self._panning_enabled = None
        self._reversed = None
        self._show_first_label = None
        self._show_last_label = None
        self._soft_max = None
        self._soft_min = None
        self._start_of_week = None
        self._start_on_tick = None
        self._tick_amount = None
        self._tick_color = None
        self._tick_interval = None
        self._tick_length = None
        self._tickmark_placement = None
        self._tick_pixel_interval = None
        self._tick_position = None
        self._tick_positioner = None
        self._tick_positions = None
        self._tick_width = None
        self._type = None
        self._unique_names = None
        self._units = None
        self._visible = None
        self._z_index = None

        self.accessibility = kwargs.get('accessibility', None)
        self.angle = kwargs.get('angle', None)
        self.ceiling = kwargs.get('ceiling', None)
        self.class_name = kwargs.get('class_name', None)
        self.crossing = kwargs.get('crossing', None)
        self.end_on_tick = kwargs.get('end_on_tick', None)
        self.events = kwargs.get('events', None)
        self.floor = kwargs.get('floor', None)
        self.grid_line_color = kwargs.get('grid_line_color', None)
        self.grid_line_dash_style = kwargs.get('grid_line_dash_style', None)
        self.grid_line_interpolation = kwargs.get('grid_line_interpolation', None)
        self.grid_line_width = kwargs.get('grid_line_width', None)
        self.grid_z_index = kwargs.get('grid_z_index', None)
        self.id = kwargs.get('id', None)
        self.labels = kwargs.get('labels', None)
        self.margin = kwargs.get('margin', None)
        self.max = kwargs.get('max', None)
        self.max_padding = kwargs.get('max_padding', None)
        self.min = kwargs.get('min', None)
        self.minor_grid_line_color = kwargs.get('minor_grid_line_color', None)
        self.minor_grid_line_dash_style = kwargs.get('minor_grid_line_dash_style', None)
        self.minor_grid_line_width = kwargs.get('minor_grid_line_width', None)
        self.minor_tick_color = kwargs.get('minor_tick_color', None)
        self.minor_tick_interval = kwargs.get('minor_tick_interval', None)
        self.minor_tick_length = kwargs.get('minor_tick_length', None)
        self.minor_tick_position = kwargs.get('minor_tick_position', None)
        self.minor_ticks = kwargs.get('minor_ticks', None)
        self.minor_tick_width = kwargs.get('minor_tick_width', None)
        self.min_padding = kwargs.get('min_padding', None)
        self.panning_enabled = kwargs.get('panning_enabled', None)
        self.reversed = kwargs.get('reversed', None)
        self.show_first_label = kwargs.get('show_first_label', None)
        self.show_last_label = kwargs.get('show_last_label', None)
        self.soft_max = kwargs.get('soft_max', None)
        self.soft_min = kwargs.get('soft_min', None)
        self.start_of_week = kwargs.get('start_of_week', None)
        self.start_on_tick = kwargs.get('start_on_tick', None)
        self.tick_amount = kwargs.get('tick_amount', None)
        self.tick_color = kwargs.get('tick_color', None)
        self.tick_interval = kwargs.get('tick_interval', None)
        self.tick_length = kwargs.get('tick_length', None)
        self.tickmark_placement = kwargs.get('tickmark_placement', None)
        self.tick_pixel_interval = kwargs.get('tick_pixel_interval', None)
        self.tick_position = kwargs.get('tick_position', None)
        self.tick_positioner = kwargs.get('tick_positioner', None)
        self.tick_positions = kwargs.get('tick_positions', None)
        self.tick_width = kwargs.get('tick_width', None)
        self.type = kwargs.get('type', None)
        self.unique_names = kwargs.get('unique_names', None)
        self.units = kwargs.get('units', None)
        self.visible = kwargs.get('visible', None)
        self.z_index = kwargs.get('z_index', None)

    @property
    def accessibility(self) -> Optional[AxisAccessibility]:
        """Accessibility options for an axis object.

        :rtype: :class:`AxisAccessibility` or :obj:`None <python:None>`
        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(AxisAccessibility)
    def accessibility(self, value):
        self._accessibility = value

    @property
    def angle(self) -> Optional[int | float | Decimal]:
        """In a polar chart, this is the angle of the Y axis in degrees, where ``0`` is
        up and ``90`` is right.  Defaults to ``0``.

        .. note::

          The angle determines the position of the axis line and the labels, though the
          coordinate system is unaffected.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = validators.numeric(value,
                                         allow_empty = True,
                                         minimum = 0)

    @property
    def ceiling(self) -> Optional[int | float | Decimal]:
        """The highest allowed value for automatically computed axis extremes. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._ceiling

    @ceiling.setter
    def ceiling(self, value):
        self._ceiling = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A class name that can then be used for styling the axis using CSS. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          The the ``class_name`` is applied to group elements for the grid, axis elements,
          and labels.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def crossing(self) -> Optional[int | float | Decimal]:
        """The value on a perpendicular axis where this axis should cross. Defaults to :obj:`None <python:None>`.

        .. tip::

          This\ris typically used on mathematical plots where the axes cross at ``0``.
          
        .. warning::
        
          When ``.crossing`` is set, space will *not* be reserved at the sides of the chart for 
          axis labels and title, so those may be clipped. In this case, it is better to place the
          axes without the ``.crossing`` option.
        
        :rtype: numeric
        """
        return self._crossing
    
    @crossing.setter
    def crossing(self, value):
        self._crossing = validators.numeric(value, allow_empty = True)

    @property
    def end_on_tick(self) -> Optional[bool]:
        """If ``True`` forces the axis to end on a tick. Defaults to ``False`` for
        :class:`XAxis`, ``True`` for :class:`YAxis`, and ``False`` for :class:`ZAxis`.

        .. hint::

          Use this option with the :meth:`GenericAxis.max_padding` setting to control the
          axis end.

        .. warning::

          This option is always disabled on a :class:`YAxis`, when panning type is either
          ``y`` or ``xy``.

        :rtype: :class:`bool <python:bool>`  or :obj:`None <python:None>`
        """
        return self._end_on_tick

    @end_on_tick.setter
    def end_on_tick(self, value):
        if value is None:
            self._end_on_tick = None
        else:
            self._end_on_tick = bool(value)

    @property
    def events(self) -> Optional[AxisEvents]:
        """Event handlers for the axis.

        :rtype: :class:`SeriesEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(AxisEvents)
    def events(self, value):
        self._events = value

    @property
    def floor(self) -> Optional[int | float | Decimal]:
        """The lowest allowed value for automatically computed axis extremes. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._floor

    @floor.setter
    def floor(self, value):
        self._floor = validators.numeric(value, allow_empty = True)

    @property
    def grid_line_color(self) -> Optional[str | Gradient | Pattern]:
        """Color of the grid lines extending the ticks across the plot area.
        Defaults to ``'#e6e6e6'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._grid_line_color

    @grid_line_color.setter
    def grid_line_color(self, value):
        from highcharts_core import utility_functions
        self._grid_line_color = utility_functions.validate_color(value)

    @property
    def grid_line_dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the grid lines. Defaults to ``Solid``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._grid_line_dash_style

    @grid_line_dash_style.setter
    def grid_line_dash_style(self, value):
        if not value:
            self._grid_line_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'grid_line_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._grid_line_dash_style = value

    @property
    def grid_line_interpolation(self) -> Optional[str]:
        """Whether the grid lines should draw as a polygon with straight lines between
        categories, or as circles. Defaults to :obj:`None <python:None>`.

        Acceptable values are:

          * ``'circle'``
          * ``'polygon'``

        .. warning::

          Only applies to :term:`polar charts <Polar Chart>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._grid_line_interpolation

    @grid_line_interpolation.setter
    def grid_line_interpolation(self, value):
        if not value:
            self._grid_line_interpolation = None
        else:
            value = validators.string(value)
            if value not in ['circle', 'polygon']:
                raise errors.HighchartsValueError(f'grid_line_interpolation expects '
                                                  f'"circle" or "polygon". Received: '
                                                  f'{value}')

            self._grid_line_interpolation = value

    @property
    def grid_line_width(self) -> Optional[int | float | Decimal]:
        """The width of the grid lines extending the ticks across the plot area. Defaults
        to ``0`` for :class:`XAxis`, ``1`` for :class:`YAxis`, and
        :obj:`None <python:None>` for :class:`ZAxis`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._grid_line_width

    @grid_line_width.setter
    def grid_line_width(self, value):
        self._grid_line_width = validators.numeric(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def grid_z_index(self) -> Optional[int | float | Decimal]:
        """The Z-index of the grid lines. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._grid_z_index

    @grid_z_index.setter
    def grid_z_index(self, value):
        self._grid_z_index = validators.numeric(value, allow_empty = True)

    @property
    def id(self) -> Optional[str]:
        """An id assigned to the axis. Defaults to :obj:`None <python:None>`.

        .. hint::

          This can be used after rendering to get a pointer to the axis object through the
          (JavaScript) ``chart.get()`` method.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = validators.string(value, allow_empty = True)

    @property
    def labels(self) -> Optional[AxisLabelOptions]:
        """Configuration settings for the axis labels, which show the number or category
        for each tick. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`AxisLabeLOptions` or :obj:`None <python:None>`
        """
        return self._labels

    @labels.setter
    @class_sensitive(AxisLabelOptions)
    def labels(self, value):
        self._labels = value

    @property
    def margin(self) -> Optional[int | float | Decimal]:
        """If there are multiple axes on the same side of the chart, the margin between
        the axes, expressed in pixels. Defaults to ``0`` for vertical axes, ``15`` for
        horizontal axes.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._margin

    @margin.setter
    def margin(self, value):
        self._margin = validators.numeric(value, allow_empty = True)

    @property
    def max(self) -> Optional[int | float | Decimal | datetime.date | datetime.datetime]:
        """The maximum value of the axis. If :obj:`None <python:None>`, the ``max`` value
        is automatically calculated. Defaults to :obj:`None <python:None>`.

        .. note::

          If the :meth:`GenericAxis.end_on_tick` is ``True``, the ``max`` value might be
          rounded up.

        .. warning::

          If a :meth:`GenericAxis.tick_amount` is set, the axis may be extended beyond the
          set ``max`` in order to reach the given number of ticks. The same may happen in
          a chart with multiple axes, determined by
          :meth:`Chart.align_ticks` where a ``tick_amount`` is applied internally.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max

    @max.setter
    def max(self, value):
        if value is None:
            self._max = None
        elif checkers.is_date(value):
            self._max = validators.date(value)
        elif checkers.is_datetime(value):
            self._max = validators.datetime(value)
        else:
            self._max = validators.numeric(value, allow_empty = True)

    @property
    def max_padding(self) -> Optional[int | float | Decimal]:
        """Padding of the max value relative to the length of the axis. Defaults to
        ``0.01``.

        For example, a value of ``0.05`` will make a 100px axis 5px longer.

        .. hint::

          This is useful when you don't want the highest data value to appear on the edge
          of the plot area.

        .. warning::

          When the :meth:`GenericAxis.max` option is set or a max extreme is set using
          (JavaScript) ``axis.setExtremes()``, the ``max_padding`` will be ignored.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_padding

    @max_padding.setter
    def max_padding(self, value):
        self._max_padding = validators.numeric(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def min(self) -> Optional[int | float | Decimal | datetime.date | datetime.datetime]:
        """The minimum value of the axis. If :obj:`None <python:None>`, the ``min`` value
        is automatically calculated. Defaults to :obj:`None <python:None>`.

        .. note::

          If the :meth:`GenericAxis.start_on_tick` is ``True``, the ``min`` value might be
          rounded down.

        .. warning::

          The automatically-calculated ``min`` value is also affected by:

            * :meth:`GenericAxis.floor`
            * :meth:`GenericAxis.soft_min`
            * :meth:`GenericAxis.min_padding`
            * :meth:`GenericAxis.min_range`
            * :meth:`GenericTypeOptions.threshold`
            * :meth:`SeriesOptions.soft_threshold`

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min

    @min.setter
    def min(self, value):
        if value is None:
            self._min = None
        elif checkers.is_date(value):
            self._min = validators.date(value)
        elif checkers.is_datetime(value):
            self._min = validators.datetime(value)
        else:
            self._min = validators.numeric(value, allow_empty = True)

    @property
    def minor_grid_line_color(self) -> Optional[str | Gradient | Pattern]:
        """Color of the minor (secondary) grid lines. Defaults to ``'#f2f2f2'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._minor_grid_line_color

    @minor_grid_line_color.setter
    def minor_grid_line_color(self, value):
        from highcharts_core import utility_functions
        self._minor_grid_line_color = utility_functions.validate_color(value)

    @property
    def minor_grid_line_dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the grid lines. Defaults to ``Solid``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._minor_grid_line_dash_style

    @minor_grid_line_dash_style.setter
    def minor_grid_line_dash_style(self, value):
        if not value:
            self._minor_grid_line_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'minor_grid_line_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._minor_grid_line_dash_style = value

    @property
    def minor_grid_line_width(self) -> Optional[int | float | Decimal]:
        """Width of the minor, secondary grid lines. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._minor_grid_line_width

    @minor_grid_line_width.setter
    def minor_grid_line_width(self, value):
        self._minor_grid_line_width = validators.numeric(value,
                                                         allow_empty = True,
                                                         minimum = 0)

    @property
    def minor_tick_color(self) -> Optional[str | Gradient | Pattern]:
        """Color for the minor tick marks. Defaults to ``'#999999'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._minor_tick_color

    @minor_tick_color.setter
    def minor_tick_color(self, value):
        from highcharts_core import utility_functions
        self._minor_tick_color = utility_functions.validate_color(value)

    @property
    def minor_tick_interval(self) -> Optional[str | int | float | Decimal]:
        """Specific tick interval in axis units for the minor ticks. Defaults to
        :obj:`None <python:None>`.

        On a linear axis, if ``"auto"``, the minor tick interval is calculated as a fifth
        of the :meth:`GenericAxis.tick_interval`. If :obj:`None <python:None>`, minor ticks
        are not shown.

        On logarithmic axes, the unit is the power of the value. For example, setting the
        ``minor_tick_interval`` to ``1`` puts one tick on each of 0.1, 1, 10, 100, etc.
        Setting the value to ``0.1`` produces 9 ticks between 1 and 10, 10 and 100 etc.

        .. warning:

          If user settings dictate minor ticks to become too dense, Highcharts will ignore
          the settings to prevent performance problems.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._minor_tick_interval

    @minor_tick_interval.setter
    def minor_tick_interval(self, value):
        if not value:
            self._minor_tick_interval = None
        else:
            if isinstance(value, str):
                value = validators.string(value)
                value = value.lower()
                if value != 'auto':
                    raise errors.HighchartsValueError(f'minor_tick_interval only accepts'
                                                      f' a string value of "auto". '
                                                      f'Received a string value of: '
                                                      f'{value}')
            else:
                value = validators.numeric(value)

            self._minor_tick_interval = value

    @property
    def minor_tick_length(self) -> Optional[int | float | Decimal]:
        """The length of the minor tick marks, in pixels. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._minor_tick_length

    @minor_tick_length.setter
    def minor_tick_length(self, value):
        self._minor_tick_length = validators.numeric(value, allow_empty = True)

    @property
    def minor_tick_position(self) -> Optional[str]:
        """The position of the minor tick marks relative to the axis line. Defaults to
        ``'outside'``.

        Accepts either:

          * ``'outside'``
          * ``'inside'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._minor_tick_position

    @minor_tick_position.setter
    def minor_tick_position(self, value):
        if not value:
            self._minor_tick_position = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['inside', 'outside']:
                raise errors.HighchartsValueError(f'minor_tick_position expects either '
                                                  f'"inside" or "outside". Received: '
                                                  f'{value}')

            self._minor_tick_position = value

    @property
    def minor_ticks(self) -> Optional[bool]:
        """Enable (``True``) or disable (``False``) minor ticks. Defaults to ``False``.

        .. note::

          Unless :meth:`GenericAxis.minor_tick_interval` is set, the minor tick interval is
          calculated as a fifth of the tickInterval.

        .. note::

          On a logarithmic axis, minor ticks are laid out based on a best guess,
          attempting to fit approximately 5 minor ticks between each major tick.

        .. warning::

          On category axes (where text is displayed in each position, rather than a
          numerical value), minor ticks are not supported.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._minor_ticks

    @minor_ticks.setter
    def minor_ticks(self, value):
        if value is None:
            self._minor_ticks = None
        else:
            self._minor_ticks = bool(value)

    @property
    def minor_tick_width(self) -> Optional[int | float | Decimal]:
        """The width of the minor tick marks, in pixels. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._minor_tick_width

    @minor_tick_width.setter
    def minor_tick_width(self, value):
        self._minor_tick_width = validators.numeric(value,
                                                    allow_empty = True,
                                                    minimum = 0)

    @property
    def min_padding(self) -> Optional[int | float | Decimal]:
        """Padding of the min value relative to the length of the axis. Defaults to
        ``0.01``.

        For example, a value of ``0.05`` will make a 100px axis 5px longer.

        .. hint::

          This is useful when you don't want the lowest data value to appear on the edge
          of the plot area.

        .. warning::

          When the :meth:`GenericAxis.min` option is set or a min extreme is set using
          (JavaScript) ``axis.setExtremes()``, the ``min_padding`` will be ignored.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_padding

    @min_padding.setter
    def min_padding(self, value):
        self._min_padding = validators.numeric(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def panning_enabled(self) -> Optional[bool]:
        """If ``True``, allows the axis to pan. ``False`` prevents the axis from panning.
        Defaults to ``True``.

        .. note::

          If :meth:`Chart.panning` is ``True`` and this option is ``False``, then this
          specific axis will not pan.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._panning_enabled

    @panning_enabled.setter
    def panning_enabled(self, value):
        if value is None:
            self._panning_enabled = None
        else:
            self._panning_enabled = bool(value)

    @property
    def reversed(self) -> Optional[bool]:
        """If ``True``, reverses the axis so that the highest number is closest to the
        origin. Defaults to :obj:`None <python:None>`.

        .. note::

          If the chart is inverted, the :class:`XAxis` is reversed by default.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reversed

    @reversed.setter
    def reversed(self, value):
        if value is None:
            self._reversed = None
        else:
            self._reversed = bool(value)

    @property
    def show_first_label(self) -> Optional[bool]:
        """If ``True``, renders the first tick label by the axis. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_first_label

    @show_first_label.setter
    def show_first_label(self, value):
        if value is None:
            self._show_first_label = None
        else:
            self._show_first_label = bool(value)

    @property
    def show_last_label(self) -> Optional[bool]:
        """If ``True``, renders the last tick label by the axis. If
        :obj:`None <python:None>`, defaults to ``True`` on
        :term:`cartesian charts` and ``False`` on
        :term:`polar charts <Polar Chart>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._show_last_label

    @show_last_label.setter
    def show_last_label(self, value):
        if value is None:
            self._show_last_label = None
        else:
            self._show_last_label = bool(value)

    @property
    def soft_max(self) -> Optional[int | float | Decimal]:
        """A soft maximum for the axis. Defaults to :obj:`None <python:None>`.

        If the series data maximum is less than this, the axis will stay at this maximum,
        but if the series data maximum is higher than this value, the axis will flex to
        show all data.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._soft_max

    @soft_max.setter
    def soft_max(self, value):
        self._soft_max = validators.numeric(value, allow_empty = True)

    @property
    def soft_min(self) -> Optional[int | float | Decimal]:
        """A soft minimum for the axis. Defaults to :obj:`None <python:None>`.

        If the series data minimum is less than this, the axis will stay at this minimum,
        but if the series data minimum is higher than this value, the axis will flex to
        show all data.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._soft_min

    @soft_min.setter
    def soft_min(self, value):
        self._soft_min = validators.numeric(value, allow_empty = True)

    @property
    def start_of_week(self) -> Optional[int]:
        """For datetime axes, this decides where to put the tick between weeks. Defaults
        to ``1`` (Monday).

        .. note::

          ``0`` = Sunday, ``1`` = Monday, etc.

        .. hint::

          As a convenience, if you supply a string with the day of week (e.g.
          ``'Monday'``), the Highcharts for Python library wlil automatically convert it
          to the appropriate numerical value.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._start_of_week

    @start_of_week.setter
    def start_of_week(self, value):
        if value is None:
            self._start_of_week = None
        else:
            if isinstance(value, str):
                value = value.lower()
                if value not in ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday',
                                 'friday', 'saturday']:
                    raise errors.HighchartsValueError(f'start_of_week expects an integer '
                                                      f'in the range 0 - 6, or a day '
                                                      f'of the week. Received: {value}')
                value = constants.DAYS_OF_WEEK.get(value)
            else:
                value = validators.integer(value, minimum = 0, maximum = 6)

            self._start_of_week = value

    @property
    def start_on_tick(self) -> Optional[bool]:
        """If ``True`` forces the axis to start on a tick. Defaults to ``False`` for
        :class:`XAxis`, ``True`` for :class:`YAxis`, and ``False`` for :class:`ZAxis`.

        .. hint::

          Use this option with the :meth:`GenericAxis.min_padding` setting to control the
          axis start.

        .. warning::

          This option is always disabled on a :class:`YAxis`, when panning type is either
          ``y`` or ``xy``.

        :rtype: :class:`bool <python:bool>`  or :obj:`None <python:None>`
        """
        return self._start_on_tick

    @start_on_tick.setter
    def start_on_tick(self, value):
        if value is None:
            self._start_on_tick = None
        else:
            self._start_on_tick = bool(value)

    @property
    def tick_amount(self) -> Optional[int]:
        """The amount of ticks to draw on the axis. Defaults to :obj:`None <python:None>`.

        .. hint::

          This provides greater control for aligning the ticks of multiple charts or panes
          within a chart.

        .. warning::

          This option overrides the :meth:`GenericAxis.tick_pixel_interval` option.

        .. note::

          This option only has an effect on linear axes. Datetime, logarithmic, or
          category axes are not affected.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._tick_amount

    @tick_amount.setter
    def tick_amount(self, value):
        self._tick_amount = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def tick_color(self) -> Optional[str | Gradient | Pattern]:
        """Color for the main tick marks. Defaults to ``'#ccd6eb'``.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._tick_color

    @tick_color.setter
    def tick_color(self, value):
        from highcharts_core import utility_functions
        self._tick_color = utility_functions.validate_color(value)

    @property
    def tick_interval(self) -> Optional[int | float | Decimal]:
        """The interval of the tick marks in axis units. Defaults to
        :obj:`None <python:None>`.

        When :obj:`None <python:None>`, the ``tick_interval`` is automatically computed
        to approximately follow :meth:`GenericAxis.tick_pixel_interval` on linear and
        datetime axes. On category axes, :obj:`None <python:None>` will default to ``1``
        (one category).

        .. note::

          Datetime axes are based on milliseconds, so for example an interval of one day
          is expressed as ``24 * 3600 * 1000``.

        On logarithmic axes, the unit is the power of the value. For example, setting the
        ``tick_interval`` to ``1`` puts one tick on each of 0.1, 1, 10, 100, etc.
        Setting the value to ``0.1`` produces 9 ticks between 1 and 10, 10 and 100 etc.

        .. warning:

          If the ``tick_interval`` is too dense for labels to be drawn, Highcharts will
          automatically remove ticks.

        .. warning::

          If the chart has multiple axes, the :meth:`GenericAxis.align_ticks` setting may
          interfere with ``tick_interval``.

        :rtype: numeric or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tick_interval

    @tick_interval.setter
    def tick_interval(self, value):
        self._tick_interval = validators.numeric(value, allow_empty = True)

    @property
    def tick_length(self) -> Optional[int | float | Decimal]:
        """The length of the main tick marks, in pixels. Defaults to ``10``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._tick_length

    @tick_length.setter
    def tick_length(self, value):
        self._tick_length = validators.numeric(value, allow_empty = True)

    @property
    def tickmark_placement(self) -> Optional[str]:
        """If ``'on'``, the tick mark is placed in the center of the category. If
        ``'between'``, the tick mark is placed between categories. If
        :obj:`None <python:None>`, defaults to ``'between'`` if
        :meth:`tick_interval <GenericAxis.tick_interval>` is ``1``, otherwise defaults to
        ``'on'``.

        .. warning::

          Applies to category axes only.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tickmark_placement

    @tickmark_placement.setter
    def tickmark_placement(self, value):
        if not value:
            self._tickmark_placement = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['on', 'between']:
                raise errors.HighchartsValueError(f'tickmark_placement expects either '
                                                  f'"on" or "between". Received: {value}')

            self._tickmark_placement = value

    @property
    def tick_pixel_interval(self) -> Optional[int | float | Decimal]:
        """If :meth:`tick_interval <GenericAxis.tick_interval>` is :obj:`None <python:None>`,
        this setting establishes the approximate interval between major tick marks,
        expressed in pixels. Defaults to ``100``.

        .. warning::

          Does not apply to categorized axes.

        .. note::

          The tick interval is also influenced by the
          :meth:`min_tick_interval <GenericAxis.min_tick_interval>` setting, which, by
          default, prevents ticks from being denser than the data points.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._tick_pixel_interval

    @tick_pixel_interval.setter
    def tick_pixel_interval(self, value):
        self._tick_pixel_interval = validators.numeric(value, allow_empty = True)

    @property
    def tick_position(self) -> Optional[str]:
        """The position of the major tick marks relative to the axis line. Defaults to
        ``'outside'``.

        Accepts either:

          * ``'outside'``
          * ``'inside'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tick_position

    @tick_position.setter
    def tick_position(self, value):
        if not value:
            self._tick_position = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['inside', 'outside']:
                raise errors.HighchartsValueError(f'tick_position expects either '
                                                  f'"inside" or "outside". Received: '
                                                  f'{value}')

            self._tick_position = value

    @property
    def tick_positioner(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function returning an array defining where the ticks are
        laid out on the axis.

        .. warning::

          This overrides the default behaviour of
          :meth:`tick_pixel_interval <GenericAxis.tick_pixel_interval>` and
          :meth:`tick_interval <GenericAxis.tick_interval>`.

        The automatic tick positions are accessible (in JavaScript) through
        ``this.tickPositions`` and can be modified by the callback.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._tick_positioner

    @tick_positioner.setter
    @class_sensitive(CallbackFunction)
    def tick_positioner(self, value):
        self._tick_positioner = value

    @property
    def tick_positions(self) -> Optional[List[int | float | Decimal]]:
        """An array that explicitly positions the major tick marks along the axis.
        Defaults to :obj:`None <python:None>`.

        .. warning::

          Setting tick positions explicitly using this setting overrides the default
          behavior of :meth:`tick_pixel_interval <GenericAxis.tick_pixel_interval>` and
          :meth:`tick_interval <GenericAxis.tick_interval>`.

        :rtype: :class:`list <python:list>` of numeric values, or
          :obj:`None <python:None>`
        """
        return self._tick_positions

    @tick_positions.setter
    def tick_positions(self, value):
        if not value:
            self._tick_positions = None
        else:
            self._tick_positions = [validators.numeric(x)
                                    for x in validators.iterable(value)]

    @property
    def tick_width(self) -> Optional[int | float | Decimal]:
        """The width of the main tick marks, in pixels. Defaults to ``0`` on category
        axes, otherwise defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._tick_width

    @tick_width.setter
    def tick_width(self, value):
        self._tick_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def type(self) -> Optional[str]:
        """The type of axis. Defaults to ``'linear'``.

        Accepts the following values:

          * ``'linear'``
          * ``'logarithmic'``
          * ``'datetime'``
          * ``'category'``

        .. note::

          In a ``'datetime'`` axis, the numbers are given in milliseconds, and tick marks
          are placed on appropriate values like full hours or days.

        .. note::

          In a ``'category'`` axis, either the :meth:`categories <GenericAxis.categories>`
          setting determines the categories rendered on the axis, or the categories are
          derived from the point names of the chart's series.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in constants.AXIS_TYPES:
                raise errors.HighchartsValueError(f'type expects a recognized  axis type.'
                                                  f'Received: {value}')

            self._type = value

    @property
    def unique_names(self) -> Optional[bool]:
        """If ``True``, points are placed on the axis according to their names. If the
        same point name is repeated in the same or another series, the point is placed on
        the same axis position as other points of the same name. When ``False``, the
        points are laid out in increasing positions regardless of their names, and the
        axis category will take the name of the last point in each position. Defaults
        to ``True``.

        .. warning::

          Applies only to category axes.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._unique_names

    @unique_names.setter
    def unique_names(self, value):
        if value is None:
            self._unique_names = None
        else:
            self._unique_names = bool(value)

    @property
    def units(self) -> Optional[List[List[str | List[int | float | Decimal | constants.EnforcedNullType | NoneType]]]]:
        """An array determining what time intervals the data is allowed to be grouped to.
        Each array item is an array where the first value is the time unit expressed as a
        :class:`str <python:str>` and the second value is another array of allowed
        multiples.

        .. warning::

          Only applies to datetime axes.

        Defaults to :obj:`None <python:None>`, which behaves as:

        .. code-block:: python

          {
              'units': [
                  [
                      'millisecond', # unit name
                      [1, 2, 5, 10, 20, 25, 50, 100, 200, 500] # allowed multiples
                  ],
                  [
                      'second',
                      [1, 2, 5, 10, 15, 30]
                  ],
                  [
                      'minute',
                      [1, 2, 5, 10, 15, 30]
                  ],
                  [
                      'hour',
                      [1, 2, 3, 4, 6, 8, 12]
                  ],
                  [
                      'day',
                      [1]
                  ],
                  [
                      'week',
                      [1]
                  ],
                  [
                      'month',
                      [1, 3, 6]
                  ],
                  [
                      'year',
                      None
                  ]
              ]
          }

        :rtype: :class:`list <python:list>` of :class:`list <python:list>` of
          :class:`str <python:str>` and :class:`list <python:list>` of numerics, or
          :obj:`None <python:None>`
        """
        return self._units

    @units.setter
    def units(self, value):
        if not value:
            self._units = None
        else:
            value = validators.iterable(value)
            value = [validators.iterable(x) for x in value]
            for item in value:
                if len(item) != 2:
                    raise errors.HighchartsValueError(f'Each entry in the units list '
                                                      f'is expected to be a 2-member '
                                                      f'list. However, was a {value}-'
                                                      f'member list.')
                validators.string(item[0])
                if item[1] and not isinstance(item[1], constants.EnforcedNullType):
                    [validators.numeric(x) for x in item[1]]

            self._units = value

    @property
    def visible(self) -> Optional[bool]:
        """If ``True``, renders the axis (including title, line, ticks, and labels). If
        ``False``, hides the axis (including title, line, ticks, and labels). Defaults to
        ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._visible

    @visible.setter
    def visible(self, value):
        if value is None:
            self._visible = None
        else:
            self._visible = bool(value)

    @property
    def z_index(self) -> Optional[int | float | Decimal]:
        """The Z-index for the axis group (including title, line, ticks, and labels).
        Defaults to ``'2'``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z_index

    @z_index.setter
    def z_index(self, value):
        self._z_index = validators.numeric(value, allow_empty = True)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.metaclasses import HighchartsMeta


class SonificationGrouping(HighchartsMeta):
    """Options for grouping data points together when sonifying. 
    
    This allows for the visual presentation to contain more points than what is being played. 
    
    .. tip::
    
      If not enabled, all visible / uncropped points are played.
      
    """
    def __init__(self, **kwargs):
        self._algorithm = None
        self._enabled = None
        self._group_timespan = None
        self._prop = None
        
        self.algorithm = kwargs.get('algorithm', None)
        self.enabled = kwargs.get('enabled', None)
        self.group_timespan = kwargs.get('group_timespan', None)
        self.prop = kwargs.get('prop', None)
        
    @property
    def algorithm(self) -> Optional[str]:
        """The grouping algorithm, which determines which points to keep when grouping a set of points together. Accepts
        ``'minmax'``, ``'first'``, ``'last'``, ``'middle'``, and ``'firstLast'``.
        
        By default ``'minmax'`` is used, which keeps both the minimum and maximum points.

        The other algorithms will either keep the first point in the group (time wise), last point, middle point, or both the first and the last point.

        The timing of the resulting point(s) is then adjusted to play evenly, regardless of its original position within the group.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._algorithm
    
    @algorithm.setter
    def algorithm(self, value):
        if not value:
            self._algorithm = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['minmax', 'first', 'last', 'middle', 'firstlast']:
                raise errors.HighchartsValueError(f'.algorithm expects either "minmax", "first", '
                                                  f'"last", "middle", or "firstLast". Received: '
                                                  f'"{value}"')
            if value == 'firstlast':
                value = 'firstLast'
            
            self._algorithm = value
            
    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, points should be grouped. Defaults to ``True``.
        
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled
    
    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)
            
    @property
    def group_timespan(self) -> Optional[int | float | Decimal]:
        """The size of each group, expressed in milliseconds. Audio events closer than this value are grouped together.
        Defaults to ``15``.
        
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._group_timespan
    
    @group_timespan.setter
    def group_timespan(self, value):
        self._group_timespan = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)
    
    @property
    def prop(self) -> Optional[str]:
        """The data point property to use when evaluating which points to keep in the group. Defaults to ``'y'``,
        which means that when the ``'minmax'`` algorithm is applied, the two points with the lowest and highest ``'y'``
        values will be kept and the others not played.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._prop
    
    @prop.setter
    def prop(self, value):
        self._prop = validators.string(value, allow_empty = True)
        
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'algorithm': as_dict.get('algorithm', None),
            'enabled': as_dict.get('enabled', None),
            'group_timespan': as_dict.get('groupTimespan', None),
            'prop': as_dict.get('prop', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'algorithm': self.algorithm,
            'enabled': self.enabled,
            'groupTimespan': self.group_timespan,
            'prop': self.prop,
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.series.data.bar import BarData
from highcharts_core.options.plot_options.bullet import TargetOptions


class BulletData(BarData):
    """Variant of :class:`BarCartesianData` which is used for data points in a bullet
    chart."""

    def __init__(self, **kwargs):
        self._target = None
        self._target_options = None

        self.target = kwargs.get('target', None)
        self.target_options = kwargs.get('target_options', None)

        super().__init__(**kwargs)

    @property
    def target(self) -> Optional[int | float | Decimal]:
        """The target value of a data point. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._target

    @target.setter
    def target(self, value):
        self._target = validators.numeric(value, allow_empty = True)

    @property
    def target_options(self) -> Optional[TargetOptions]:
        """Options related to the look and positioning of the
        :meth:`target <BulletData.target>`.

        :rtype: :class:`TargetOptions` or :obj:`None <python:None>`
        """
        return self._target_options

    @target_options.setter
    @class_sensitive(TargetOptions)
    def target_options(self, value):
        self._target_options = value

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'BulletData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls()
            elif checkers.is_iterable(item):
                if len(item) == 3:
                    as_dict = {
                        'x': item[0],
                        'y': item[1],
                        'target': item[2]
                    }
                elif len(item) == 2:
                    as_dict = {
                        'x': None,
                        'y': item[0],
                        'target': item[1]
                    }
                else:
                    raise errors.HighchartsValueError(f'data expects either a 3D or 2D '
                                                      f'collection. Collection received '
                                                      f'had {len(item)} dimensions.')
                    
                as_obj = cls.from_dict(as_dict)
                if checkers.is_string(as_obj.x):
                    as_obj.name = as_obj.x
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be a Bullet Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank', None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'data_labels': as_dict.get('dataLabels', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'drilldown': as_dict.get('drilldown', None),
            'marker': as_dict.get('marker', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),

            'border_color': as_dict.get('borderColor', None),
            'border_width': as_dict.get('borderWidth', None),
            'dash_style': as_dict.get('dashStyle', None),
            'point_width': as_dict.get('pointWidth', None),

            'target': as_dict.get('target', None),
            'target_options': as_dict.get('targetOptions', None),

        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'target': self.target,
            'targetOptions': self.target_options
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators, checkers

from highcharts_core import constants, errors
from highcharts_core.decorators import validate_types
from highcharts_core.utility_classes import Gradient, Pattern
from highcharts_core.options.annotations.points import AnnotationPoint
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.utility_functions import validate_color


class ShapeOptionsBase(HighchartsMeta):
    """Base set of options applied to all annotation shapes."""

    def __init__(self, **kwargs):
        self._dash_style = None
        self._fill = None
        self._height = None
        self._ry = None
        self._snap = None
        self._src = None
        self._stroke = None
        self._stroke_width = None
        self._x_axis = None
        self._y_axis = None

        self.dash_style = kwargs.get('dash_style', None)
        self.fill = kwargs.get('fill', None)
        self.height = kwargs.get('height', None)
        self.ry = kwargs.get('ry', None)
        self.snap = kwargs.get('snap', None)
        self.src = kwargs.get('src', None)
        self.stroke = kwargs.get('stroke', None)
        self.stroke_width = kwargs.get('stroke_width', None)
        self.x_axis = kwargs.get('x_axis', None)
        self.y_axis = kwargs.get('y_axis', None)

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the shape's stroke.

        Accepts the following values:

          * ``'Solid'``
          * ``'ShortDash'``
          * ``'ShortDot'``
          * ``'ShortDashDot'``
          * ``'ShortDashDotDot'``
          * ``'Dot'``
          * ``'Dash'``
          * ``'LongDash'``
          * ``'DashDot'``
          * ``'LongDashDot'``
          * ``'LongDashDotDot'``

        :returns: The name of the dash style to apply to the shape's stroke.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._dash_style = None
        else:
            if value not in constants.SHAPES_ALLOWED_DASH_STYLES:
                raise errors.HighchartsValueError(f'dash_style expects a supported value.'
                                                  f' Received: {value}')
            self._dash_style = value

    @property
    def fill(self) -> Optional[str | Gradient | Pattern]:
        """The color of the shape's fill. Defaults to ``'rgba(0, 0, 0, 0.75)'``.

        :rtype: :class:`str <python:str>` (for colors), :class:`Gradient` for gradients,
          :class:`Pattern` for pattern definitions, or :obj:`None <python:None>`
        """
        return self._fill

    @fill.setter
    def fill(self, value):
        self._fill = validate_color(value)

    @property
    def height(self) -> Optional[int | float | Decimal]:
        """The height of the shape in pixels.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = validators.numeric(value, allow_empty = True)

    @property
    def r(self) -> Optional[int | float | Decimal]:
        """The radius of the shape in pixels. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._r

    @r.setter
    def r(self, value):
        self._r = validators.numeric(value, allow_empty = True)

    @property
    def ry(self) -> Optional[int | float | Decimal]:
        """The radius of the shape along the vertical dimension. Used to draw ellipses.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._ry

    @ry.setter
    def ry(self, value):
        self._ry = validators.numeric(value, allow_empty = True)

    @property
    def snap(self) -> Optional[int | float | Decimal]:
        """Defines additional snapping area around an annotation making this annotation
        to focus. Defined in pixels.

        Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._snap

    @snap.setter
    def snap(self, value):
        self._snap = validators.numeric(value, allow_empty = True)

    @property
    def src(self) -> Optional[str]:
        """The URL for an image to use as the annotation shape.

        .. note::

          :meth:`ShapeOptions.type` has to be set to ``'image'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises HighchartsValueError: if a value is supplied that is not a URL or not
          path-like

        """
        return self._src

    @src.setter
    def src(self, value):
        if not value:
            self._src = None
        else:
            try:
                self._src = validators.url(value)
            except ValueError:
                try:
                    self._src = validators.path(value)
                except ValueError:
                    raise errors.HighchartsValueError(f'value provided ({value}) not a '
                                                      f'valid URL or path')

    @property
    def stroke(self) -> Optional[str]:
        """The color of the shape's stroke. Defaults to
        ``'rgba(0, 0, 0, 0.75)'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stroke

    @stroke.setter
    def stroke(self, value):
        self._stroke = validators.string(value, allow_empty = True)

    @property
    def stroke_width(self) -> Optional[int | float | Decimal]:
        """The pixel stroke width of the shape. Defaults to
        ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._stroke_width

    @stroke_width.setter
    def stroke_width(self, value):
        self._stroke_width = validators.numeric(value, allow_empty = True)

    @property
    def type(self) -> Optional[str]:
        """The type of the shape. Defaults to ``'rect'``.

        Accepts:

          * ``'rect'``
          * ``'circle'``
          * ``'ellipse'``
          * ``'image'``

        :returns: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises HighchartsValueError: if type not supported
        """
        return self._type

    @type.setter
    def type(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._type = None
        else:
            value = value.lower()
            if value not in ['rect', 'circle', 'ellipse']:
                raise errors.HighchartsValueError(f'ShapeOptions.type accepts either '
                                                  f'"rect", "circle", "image", or '
                                                  f'"ellipse". Received: {value}')
            self._type = value

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width of the shape, expressed in pixels.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @property
    def x_axis(self) -> Optional[int]:
        """The X-Axis index to which the points should be attached.

        .. note::

          Used for the ``ellipse`` :meth:`type <ShapeOptions.type>`

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if set to a negative value

        """
        return self._x_axis

    @x_axis.setter
    def x_axis(self, value):
        self._x_axis = validators.integer(value,
                                          allow_empty = True,
                                          minimum = 0,
                                          coerce_value = True)

    @property
    def y_axis(self) -> Optional[int]:
        """The Y-Axis index to which the points should be attached.

        .. note::

          Used for the ``ellipse`` :meth:`type <ShapeOptions.type>`

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if set to a negative value
        """
        return self._y_axis

    @y_axis.setter
    def y_axis(self, value):
        self._y_axis = validators.integer(value,
                                          allow_empty = True,
                                          minimum = 0,
                                          coerce_value = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'dash_style': as_dict.get('dashStyle', None),
            'fill': as_dict.get('fill', None),
            'height': as_dict.get('height', None),
            'ry': as_dict.get('ry', None),
            'snap': as_dict.get('snap', None),
            'src': as_dict.get('src', None),
            'stroke': as_dict.get('stroke', None),
            'stroke_width': as_dict.get('strokeWidth', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dashStyle': self.dash_style,
            'fill': self.fill,
            'height': self.height,
            'ry': self.ry,
            'snap': self.snap,
            'src': self.src,
            'stroke': self.stroke,
            'strokeWidth': self.stroke_width,
            'xAxis': self.x_axis,
            'yAxis': self.y_axis
        }

        return untrimmed


class ShapeOptions(ShapeOptionsBase):
    """Global options applied to all annotation shapes."""

    def __init__(self, **kwargs):
        self._r = None
        self._type = None
        self._width = None

        self.r = kwargs.get('r', None)
        self.type = kwargs.get('type', None)
        self.width = kwargs.get('width', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.shapeOptions'

    @property
    def r(self) -> Optional[int | float | Decimal]:
        f"""The radius of the shape in pixels. Defaults to {constants.DEFAULT_SHAPES_R}.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._r

    @r.setter
    def r(self, value):
        self._r = validators.numeric(value, allow_empty = True)

    @property
    def type(self) -> Optional[str]:
        f"""The type of the shape. Defaults to ``{constants.DEFAULT_SHAPES_TYPE}``.

        Accepts:

          * ``'rect'``
          * ``'circle'``
          * ``'ellipse'``
          * ``'image'``

        :returns: :class:`str <python:str>` or :obj:`None <python:None>`

        :raises HighchartsValueError: if type not supported
        """
        return self._type

    @type.setter
    def type(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._type = None
        else:
            value = value.lower()
            if value not in ['rect', 'circle', 'ellipse']:
                raise errors.HighchartsValueError(f'ShapeOptions.type accepts either '
                                                  f'"rect", "circle", "image", or '
                                                  f'"ellipse". Received: {value}')
            self._type = value

    @property
    def width(self) -> Optional[int | float | Decimal]:
        """The width of the shape, expressed in pixels.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        self._width = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'dash_style': as_dict.get('dashStyle', None),
            'fill': as_dict.get('fill', None),
            'height': as_dict.get('height', None),
            'r': as_dict.get('r', None),
            'ry': as_dict.get('ry', None),
            'snap': as_dict.get('snap', None),
            'src': as_dict.get('src', None),
            'stroke': as_dict.get('stroke', None),
            'stroke_width': as_dict.get('strokeWidth', None),
            'type': as_dict.get('type', None),
            'width': as_dict.get('width', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dashStyle': self.dash_style,
            'fill': self.fill,
            'height': self.height,
            'r': self.r,
            'ry': self.ry,
            'snap': self.snap,
            'src': self.src,
            'stroke': self.stroke,
            'strokeWidth': self.stroke_width,
            'type': self.type,
            'width': self.width,
            'xAxis': self.x_axis,
            'yAxis': self.y_axis
        }

        return untrimmed


class AnnotationShape(ShapeOptions):
    """Configuration for an annotation shape applied to a specific point.

    Used to override the global settings configured in :class:`ShapeOptions` and applied
    via :meth:`Annotation.shape_options`.
    """

    def __init__(self, **kwargs):
        self._marker_end = None
        self._marker_start = None
        self._point = None
        self._points = None

        self.marker_end = kwargs.get('marker_end', None)
        self.marker_start = kwargs.get('marker_start', None)
        self.point = kwargs.get('point', None)
        self.points = kwargs.get('points', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.shapes'

    @property
    def marker_end(self) -> Optional[str]:
        """ID of the marker which will be drawn at the final vertex of the path.

        .. note::

          Custom markers can be defined in the :meth:`Options.defs` property.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._marker_end

    @marker_end.setter
    def marker_end(self, value):
        self._marker_end = validators.string(value, allow_empty = True)

    @property
    def marker_start(self) -> Optional[str]:
        """ID of the marker which will be drawn at the first vertex of the path.

        .. note::

          Custom markers can be defined in the :meth:`Options.defs` property.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._marker_start

    @marker_start.setter
    def marker_start(self, value):
        self._marker_start = validators.string(value, allow_empty = True)

    @property
    def point(self) -> Optional[str | AnnotationPoint]:
        """Determines the point to which the shape will be connected.

        It can be either the ID of the point which exists in the series, or a new point
        with defined x, y properties and optionally axes.

        :rtype: :class:`str <python:str>` or :class:`AnnotationPoint` or
          :obj:`None <python:None>`

        :raises HighchartsValueError: if cannot resolve the value to an allowed type
        """
        return self._point

    @point.setter
    def point(self, value):
        if not value:
            self._point = None
        elif isinstance(value, AnnotationPoint):
            self._point = value
        elif isinstance(value, str):
            try:
                self._point = AnnotationPoint.from_json(value)
            except ValueError:
                self._point = validators.string(value)
        elif isinstance(value, dict):
            self._point = AnnotationPoint.from_dict(value)
        else:
            raise errors.HighchartsValueError('Unable to resolve the value supplied to a '
                                              'supported type.')

    @property
    def points(self) -> Optional[List[CallbackFunction | AnnotationPoint | str]]:
        """An array of points for the shape or a JavaScript callback function that returns
        that shape point.

        .. note::

          This option is available for shapes which can use multiple points such as path.
          A point can be either a :class:`AnnotationPoint` object or a point's id.

        :rtype: :class:`list <python:list>` of :class:`CallbackFunction` or
          :class:`str <python:str>` or :class:`AnnotationPoint` or
          :class:`str <python:str>`, OR :obj:`None <python:None>`
        """
        return self._points

    @points.setter
    def points(self, value):
        if not value:
            self._points = None
        elif checkers.is_iterable(value):
            processed_value = []
            for item in value:
                try:
                    item = validate_types(item, types = AnnotationPoint)
                except ValueError:
                    try:
                        item = validate_types(item, types = CallbackFunction)
                    except (ValueError, TypeError):
                        try:
                            item = validators.string(item, allow_empty = False)
                        except ValueError:
                            raise errors.HighchartsValueError(
                                f'points expects '
                                f'AnnotationPoint, '
                                f'CallbackFunction, or str'
                                f'instances. Received: '
                                f'{item.__class__.__name__}'
                            )
                processed_value.append(item)

            self._points = [x for x in processed_value]
        else:
            raise errors.HighchartsValueError('Unable to resolve the value to a '
                                              'supported type.')

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            # from ShapeOptions
            'dash_style': as_dict.get('dashStyle', None),
            'fill': as_dict.get('fill', None),
            'height': as_dict.get('height', None),
            'r': as_dict.get('r', None),
            'ry': as_dict.get('ry', None),
            'snap': as_dict.get('snap', None),
            'src': as_dict.get('src', None),
            'stroke': as_dict.get('stroke', None),
            'stroke_width': as_dict.get('strokeWidth', None),
            'type': as_dict.get('type', None),
            'width': as_dict.get('width', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),

            # from AnnotationShape
            'marker_end': as_dict.get('markerEnd', None),
            'marker_start': as_dict.get('markerStart', None),
            'point': as_dict.get('point', None),
            'points': as_dict.get('points', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'markerEnd': self.marker_end,
            'markerStart': self.marker_start,
            'point': self.point,
            'points': self.points,
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


__all__ = [
    'ShapeOptions',
    'AnnotationShape',
    'ShapeOptionsBase'
]



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.chart.reset_zoom_button import ResetZoomButtonOptions

from highcharts_core import errors


class MouseWheelOptions(HighchartsMeta):
    """Configuration of zooming via the mouse wheel."""
    
    def __init__(self, **kwargs):
        self._enabled = None
        self._sensitivity = None
        self._type = None
        
        self.enabled = kwargs.get('enabled', None)
        self.sensitivity = kwargs.get('sensitivity', None)
        self.type = kwargs.get('type', None)
        
    @property
    def enabled(self) -> Optional[bool]:
        """Enables or disables zooming via the mouse wheel. Defaults to ``True``.
        
        :rtype: :class:`bool <python:bool>`
        """
        return self._enabled
    
    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)
            
    @property
    def sensitivity(self) -> Optional[int | float | Decimal]:
        """Adjust the sensitivity of mouse wheel when controlling zoom. 
        
        A value of ``1`` is no sensitivity, while a value of ``2``, one mouse wheel 
        rotation will in ``50%``.
        
        Defaults to ``1.1``.
        
        :rtype: numeric
        """
        return self._sensitivity
    
    @sensitivity.setter
    def sensitivity(self, value):
        self._sensitivity = validators.numeric(value, allow_empty = True)
        
    @property
    def type(self) -> Optional[str]:
        """Decides in what dimensions the user can zoom scrolling the mouse wheel.
        Accepts one of ``'x'``, ``'y'`` or ``'xy'``. 
        
        If :obj:`None <python:None>`, it will inherit from
        :meth:`ZoomingOptions.type <highcharts_core.options.chart.zooming.ZoomingOptions.type>`.
        
        .. warning::
        
          Note that particularly when zooming via the mouse wheel in the ``'y'`` direction, 
          the zoom is affected by the default 
          :meth:`YAxisOptions.start_on_tick <highcharts_core.options.axes.YAxisOptions.start_on_tick>`
          and
          :meth:`YAxisOptions.end_on_tick <highcharts_core.options.axes.YAxisOptions.end_on_tick>`
          settings. 
          
          In order to respect these settings, the zoom level will adjust after the user has stopped 
          zooming. To prevent this, consider setting both those properties to ``False``.
          
        :rtype: :class:`str <python:str>`
        '"""
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['x', 'y', 'xy']:
                raise errors.HighchartsValueError(f'MouseWheelOptions.type expects a value '
                                                  f'of "x", "y", or "xy". Received: "{value}".')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'enabled': as_dict.get('enabled', None),
            'sensitivity': as_dict.get('sensitivity', None),
            'type': as_dict.get('type', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'enabled': self.enabled,
            'sensitivity': self.sensitivity,
            'type': self.type,
        }

        return untrimmed


class ZoomingOptions(HighchartsMeta):
    """Chart zooming configuration."""

    def __init__(self, **kwargs):
        self._key = None
        self._mouse_wheel = None
        self._pinch_type = None
        self._reset_button = None
        self._single_touch = None
        self._type = None

        self.key = kwargs.get('key', None)
        self.mouse_wheel = kwargs.get('mouse_wheel', None)
        self.pinch_type = kwargs.get('pinch_type', None)
        self.reset_button = kwargs.get('reset_button', None)
        self.single_touch = kwargs.get('single_touch', None)
        self.type = kwargs.get('type', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'chart.zooming'

    @property
    def key(self) -> Optional[str]:
        """Sets a key to hold when dragging to zoom the chart.

        .. hint::

          This is useful to avoid zooming while moving points.

        .. hint::

          This should be set to a different key than
          :meth:`ChartOptions.pan_key <highcharts_core.options.chart.ChartOptions.pan_key>`.

        Accepts the following values:

          * ``'alt'``
          * ``'ctrl'``
          * ``'meta'`` (the command key on Mac and Windows key on Windows)
          * ``'shift'``.

        The keys are mapped directly to the key properties of the click event argument
        (``event.altKey``, ``event.ctrlKey``, ``event.metaKey``, and ``event.shiftKey``).

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._key

    @key.setter
    def key(self, value):
        if not value:
            self._key = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['alt', 'ctrl', 'meta', 'shift']:
                raise errors.HighchartsValueError(f'pan_key expects a value of "alt", '
                                                  f'"ctrl", "meta", or "shift". Was: '
                                                  f'{value}')

            self._key = value

    @property
    def mouse_wheel(self) -> Optional[bool | MouseWheelOptions]:
        """Configuration of how users can use the mouse wheel to control zooming within the chart.
        
        .. note::
        
          Zooming via the mouse wheel is enabled by default. To disable, set this option to ``False``.
        
        .. note::
        
          Zooming via the mouse wheel is a feature included in 
          `Highcharts Stock <https://highcharts.com/products/stock>`__, but is also available 
          in `Highcharts Core <https://highcharts.com/products/highcharts>`__ by loading
          ``modules/mouse-wheel-zoom.js``.
          
        :rtype: :class:`bool <python:bool>`
        """
        return self._mouse_wheel
    
    @mouse_wheel.setter
    def mouse_wheel(self, value):
        if value is None:
            self._mouse_wheel = None
        else:
            try:
                value = validate_types(value, MouseWheelOptions)
            except (ValueError, TypeError):
                value = bool(value)
                
            if value is True:
                value = MouseWheelOptions(enabled = True)
                
            self._mouse_wheel = value

    @property
    def pinch_type(self) -> Optional[str]:
        """Equivalent to :meth:`.type <ZoomingOptions.type>` but for multi-touch gestures
        only. Defaults to ``'x'``.

        Accepts:

          * ``'x'``
          * ``'y'``
          * ``'xy'``
          * :obj:`None <python:None>`

        If not specified explicitly, the pinch type is the same as the
        :meth:`.type <ZoomingOptions.type>`. However, pinching can be enabled separately
        in some cases, for example in stock charts where a mouse drag pans the chart,
        while pinching is enabled. When :meth:`Tooltip.follow_touch_move` is ``True``,
        pinch type only applies to two-finger touches.

        :returns: The configuration of the pinch directional support.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._pinch_type

    @pinch_type.setter
    def pinch_type(self, value):
        if not value:
            self._pinch_type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['x', 'y', 'xy']:
                raise errors.HighchartsValueError(f'pinch_type is expected to be either '
                                                  f'"x", "y", "xy", or None. Was: '
                                                  f'{value}')
            self._pinch_type = value

    @property
    def reset_button(self) -> Optional[ResetZoomButtonOptions]:
        """Configuration settings for the button that appears after a selection zoom,
        allowing the user to reset zoom.

        :rtype: :class:`ResetZoomButtonOptions` or :obj:`None <python:None>`
        """
        return self._reset_button

    @reset_button.setter
    @class_sensitive(ResetZoomButtonOptions)
    def reset_button(self, value):
        self._reset_button = value

    @property
    def single_touch(self) -> Optional[bool]:
        """If ``True``, enables zooming with a single touch (in combination with
        :meth:`.type <ZoomingOptions.type>`) while two-finger pinch will still work as per
        :meth:`.pinch_type <ZoomingOptions.pinch_type>`. Defaults to ``False``.

        .. warning::

          Enabling zoom by single touch will interfere with touch-dragging the chart to
          read the tooltip, and if vertical zooming is enabled will make it hard to scroll
          vertically on the page.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._single_touch

    @single_touch.setter
    def single_touch(self, value):
        if value is None:
            self._single_touch = None
        else:
            self._single_touch = bool(value)

    @property
    def type(self) -> Optional[str]:
        """Determines in which dimensions the user can zoom by dragging the mouse. By
        default, not set.

        Accepts:

          * ``'x'``
          * ``'y'``
          * ``'xy'``
          * :obj:`None <python:None>`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._type

    @type.setter
    def type(self, value):
        if not value:
            self._type = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['x', 'y', 'xy']:
                raise errors.HighchartsValueError(f'type is expected to be either '
                                                  f'"x", "y", "xy", or None. Was: '
                                                  f'{value}')
            self._type = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'key': as_dict.get('key', None),
            'mouse_wheel': as_dict.get('mouseWheel', None),
            'pinch_type': as_dict.get('pinchType', None),
            'reset_button': as_dict.get('resetButton', None),
            'single_touch': as_dict.get('singleTouch', None),
            'type': as_dict.get('type', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'key': self.key,
            'mouseWheel': self.mouse_wheel,
            'pinchType': self.pinch_type,
            'resetButton': self.reset_button,
            'singleTouch': self.single_touch,
            'type': self.type
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.plot_options.bar import BarOptions
from highcharts_core.options.plot_options.drag_drop import BulletDragDropOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class TargetOptions(HighchartsMeta):
    """All options related to the look and positiong of targets."""

    def __init__(self, **kwargs):
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._color = None
        self._width = None
        self._height = None

        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.color = kwargs.get('color', None)
        self.height = kwargs.get('height', None)
        self.width = kwargs.get('width', None)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The border color of the rectangle representing the target. When
        :obj:`None <python:None>`, the point's border color is used. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The corner radius of the rectangle surrounding the target. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width of the rectangle representing the target. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the rectangle representing the target. When
        :obj:`None <python:None>`, applies the point's color (if set in point's options -
        ``color``), or the zone of the target value (if ``zones`` or ``negative_color``
        are set), or the same color as the point.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def height(self) -> Optional[int | float | Decimal]:
        """The height of the rectangle representing the target, expressed in pixels.
        Defaults to ``3``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._height

    @height.setter
    def height(self, value):
        self._height = validators.numeric(value, allow_empty = True)

    @property
    def width(self) -> Optional[str | int | float | Decimal]:
        """The width of the rectangle representing the target. Could be set as a pixel
        value or as a percentage of a column width. Defaults to ``'140%'``.

        :rtype: numeric, :class:`str <python:str>`, or :obj:`None <python:None>`
        """
        return self._width

    @width.setter
    def width(self, value):
        if value is None:
            self._width = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('width expects either a number or '
                                                      'a % string. No "%" character '
                                                      'found.')
            except TypeError:
                value = validators.numeric(value, minimum = 0)

            self._width = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'color': as_dict.get('color', None),
            'height': as_dict.get('height', None),
            'width': as_dict.get('width', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'color': self.color,
            'height': self.height,
            'width': self.width
        }

        return untrimmed


class BulletOptions(BarOptions):
    """General options to apply to all Bullet series types.

    A bullet graph is a variation of a bar graph. The bullet graph features a single
    measure, compares it to a target, and displays it in the context of qualitative
    ranges of performance that could be set using :meth:`YAxis.plot_bands`.

    .. figure:: ../../../_static/bullet-example.png
      :alt: Bullet Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._target_options = None

        self.target_options = kwargs.get('target_options', None)

        super().__init__(**kwargs)

    @property
    def drag_drop(self) -> Optional[BulletDragDropOptions]:
        """The draggable-points module allows points to be moved around or modified in the
        chart.

        In addition to the options mentioned under the dragDrop API structure, the module
        fires three (JavaScript) events:

          * ``point.dragStart``
          * ``point.drag``
          * ``point.drop``

        :rtype: :class:`DragDropOptions` or :obj:`None <python:None>`
        """
        return self._drag_drop

    @drag_drop.setter
    @class_sensitive(BulletDragDropOptions)
    def drag_drop(self, value):
        self._drag_drop = value

    @property
    def target_options(self) -> Optional[TargetOptions]:
        """All options related to the look and positiong of targets.

        :rtype: :class:`TargetOptions` or :obj:`None <python:None>`
        """
        return self._target_options

    @target_options.setter
    @class_sensitive(TargetOptions)
    def target_options(self, value):
        self._target_options = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'center_in_category': as_dict.get('centerInCategory', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'colors': as_dict.get('colors', None),
            'grouping': as_dict.get('grouping', None),
            'group_padding': as_dict.get('groupPadding', None),
            'max_point_width': as_dict.get('maxPointWidth', None),
            'min_point_length': as_dict.get('minPointLength', None),
            'point_padding': as_dict.get('pointPadding', None),
            'point_range': as_dict.get('pointRange', None),
            'point_width': as_dict.get('pointWidth', None),

            'depth': as_dict.get('depth', None),
            'edge_color': as_dict.get('edgeColor', None),
            'edge_width': as_dict.get('edgeWidth', None),
            'group_z_padding': as_dict.get('groupZPadding', None),

            'target_options': as_dict.get('targetOptions', None)

        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'targetOptions': self.target_options
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.options.plot_options.drag_drop import DragDropOptions
from highcharts_core.utility_classes.shadows import ShadowOptions


class TimelineOptions(GenericTypeOptions):
    """General options to apply to all Timeline series types.

    The timeline series presents given events along a drawn line.

    .. tabs::

      .. tab:: Standard Timeline

        .. figure:: ../../../_static/timeline-example.png
          :alt: Timeline Example Chart
          :align: center

      .. tab:: Inverted Timeline

        .. figure:: ../../../_static/timeline-example-inverted.png
          :alt: Inverted Timeline Example Chart
          :align: center

      .. tab:: With True Datetime Axis

        .. figure:: ../../../_static/timeline-example-datetime.png
          :alt: Timeline Example Chart with Datetime Axis
          :align: center

    """

    def __init__(self, **kwargs):
        self._color_axis = None
        self._color_by_point = None
        self._color_index = None
        self._color_key = None
        self._crisp = True
        self._drag_drop = None
        self._ignore_hidden_point = None
        self._linecap = None
        self._relative_x_value = None
        self._shadow = None

        self.color_axis = kwargs.get('color_axis', None)
        self.color_by_point = kwargs.get('color_by_point', None)
        self.color_index = kwargs.get('color_index', None)
        self.color_key = kwargs.get('color_key', None)
        self.crisp = kwargs.get('crisp', None)
        self.drag_drop = kwargs.get('drag_drop', None)
        self.ignore_hidden_point = kwargs.get('ignore_hidden_point', None)
        self.linecap = kwargs.get('linecap', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)
        self.shadow = kwargs.get('shadow', None)

        super().__init__(**kwargs)

    @property
    def color_axis(self) -> Optional[str | int | bool]:
        """When using dual or multiple color axes, this setting defines which
        :term:`color axis` the particular series is connected to. It refers to either the
        :meth:`ColorAxis.id` or the index of the axis in the :class:`ColorAxis` array,
        with ``0`` being the first. Set this option to ``False`` to prevent a series from
        connecting to the default color axis.

        Defaults to ``0``.

        :rtype: :obj:`None <python:None>` or :class:`str <python:str>` or
          :class:`int <python:int>` or :class:`bool <python:bool>`
        """
        return self._color_axis

    @color_axis.setter
    def color_axis(self, value):
        if value is None:
            self._color_axis = None
        elif value is False:
            self._color_axis = False
        else:
            try:
                self._color_axis = validators.string(value)
            except TypeError:
                self._color_axis = validators.integer(value,
                                                      minimum = 0)

    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def color_key(self) -> Optional[str]:
        """Determines what data value should be used to calculate point color if
        :meth:`AreaOptions.color_axis` is used.

        .. note::

          Requires to set ``min`` and ``max`` if some custom point property is used or if
          approximation for data grouping is set to ``'sum'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color_key

    @color_key.setter
    def color_key(self, value):
        self._color_key = validators.string(value, allow_empty = True)

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def drag_drop(self) -> Optional[DragDropOptions]:
        """The draggable-points module allows points to be moved around or modified in the
        chart.

        In addition to the options mentioned under the dragDrop API structure, the module
        fires three (JavaScript) events:

          * ``point.dragStart``
          * ``point.drag``
          * ``point.drop``

        :rtype: :class:`DragDropOptions` or :obj:`None <python:None>`
        """
        return self._drag_drop

    @drag_drop.setter
    @class_sensitive(DragDropOptions)
    def drag_drop(self, value):
        self._drag_drop = value

    @property
    def ignore_hidden_point(self) -> Optional[bool]:
        """If ``True``, the series shall be redrawn as if the hidden points were ``null``.
        If ``False``, hidden points will not be displayed but the slice will still be
        drawn as a gap in the pie.

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._ignore_hidden_point

    @ignore_hidden_point.setter
    def ignore_hidden_point(self, value):
        if value is None:
            self._ignore_hidden_point = None
        else:
            self._ignore_hidden_point = bool(value)

    @property
    def linecap(self) -> Optional[str]:
        """The SVG value used for the ``stroke-linecap`` and ``stroke-linejoin`` of a line
        graph. Defaults to ``'round'``, which means that lines are rounded in the ends and
        bends.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._linecap

    @linecap.setter
    def linecap(self, value):
        self._linecap = validators.string(value, allow_empty = True)

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the tooltip. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or a boolean setting which hides the
          shadow or displays the default shadow.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if isinstance(value, bool):
            self._shadow = value
        elif not value:
            self._shadow = None
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'color_axis': as_dict.get('colorAxis', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'crisp': as_dict.get('crisp', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'ignore_hidden_point': as_dict.get('ignoreHiddenPoint', None),
            'linecap': as_dict.get('linecap', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'colorAxis': self.color_axis,
            'colorByPoint': self.color_by_point,
            'colorIndex': self.color_index,
            'colorKey': self.color_key,
            'crisp': self.crisp,
            'dragDrop': self.drag_drop,
            'ignoreHiddenPoint': self.ignore_hidden_point,
            'linecap': self.linecap,
            'relativeXValue': self.relative_x_value,
            'shadow': self.shadow
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.plot_options import PlotOptions as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
]

@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('plot_options/plot_options/01.js', False, None),

    ('plot_options/plot_options/01.js', True, None),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class RangeSelectorLanguageOptions(HighchartsMeta):
    """Language options for range selectors when used in accessibility mode."""

    def __init__(self, **kwargs):
        self._click_button_announcement = None
        self._dropdown_label = None
        self._max_input_label = None
        self._min_input_label = None

        self.click_button_announcement = kwargs.get('click_button_announcement', None)
        self.dropdown_label = kwargs.get('dropdown_label', None)
        self.max_input_label = kwargs.get('max_input_label', None)
        self.min_input_label = kwargs.get('min_input_label', None)

    @property
    def click_button_announcement(self) -> Optional[str]:
        """Defaults to
        ``'Viewing {axisRangeDescription}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._click_button_announcement

    @click_button_announcement.setter
    def click_button_announcement(self, value):
        self._click_button_announcement = validators.string(value, allow_empty = True)

    @property
    def dropdown_label(self) -> Optional[str]:
        """Defaults to
        ``'{rangeTitle}'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dropdown_label

    @dropdown_label.setter
    def dropdown_label(self, value):
        self._dropdown_label = validators.string(value, allow_empty = True)

    @property
    def max_input_label(self) -> Optional[str]:
        """Defaults to
        ``'Select end date.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._max_input_label

    @max_input_label.setter
    def max_input_label(self, value):
        self._max_input_label = validators.string(value, allow_empty = True)

    @property
    def min_input_label(self) -> Optional[str]:
        """Defaults to
        ``'Select start date.'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._min_input_label

    @min_input_label.setter
    def min_input_label(self, value):
        self._min_input_label = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'click_button_announcement': as_dict.get('clickButtonAnnouncement', None),
            'dropdown_label': as_dict.get('dropdownLabel', None),
            'max_input_label': as_dict.get('maxInputLabel', None),
            'min_input_label': as_dict.get('minInputLabel', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'clickButtonAnnouncement': self.click_button_announcement,
            'dropdownLabel': self.dropdown_label,
            'maxInputLabel': self.max_input_label,
            'minInputLabel': self.min_input_label
        }

        return untrimmed



"""Tests for ``highcharts.tooltips``."""

import pytest

from validator_collection import checkers

from highcharts_core.options.tooltips import Tooltip as cls
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'animation': True,
      'background_color': '#ccc',
      'border_color': '#999',
      'border_radius': 4,
      'border_width': 1,
      'class_name': 'some-class-name',
      'cluster_format': 'format string',
      'date_time_label_formats': {
        'day': 'test',
        'hour': 'test',
        'millisecond': 'test',
        'minute': 'test',
        'month': 'test',
        'second': 'test',
        'week': 'test',
        'year': 'test'
      },
      'distance': 12,
      'enabled': True,
      'follow_pointer': True,
      'follow_touch_move': True,
      'footer_format': 'format string',
      'formatter': """function() { return true; }""",
      'header_format': 'format string',
      'header_shape': 'circle',
      'hide_delay': 3,
      'null_format': 'format string',
      'null_formatter': """function() { return true; }""",
      'outside': False,
      'padding': 6,
      'point_format': 'format string',
      'point_formatter': """function() { return true; }""",
      'positioner': """function() { return true; }""",
      'shadow': False,
      'shape': 'rect',
      'shared': False,
      'snap': 4,
      'split': False,
      'stick_on_contact': True,
      'style': 'style string goes here',
      'use_html': False,
      'value_decimals': 2,
      'value_prefix': '$',
      'value_suffix': ' USD',
      'x_date_format': 'format string'
     }, None),
    ({
        'formatter': CallbackFunction.from_js_literal("""function () {
            // The first returned item is the header, subsequent items are the
            // points
            return ['<b>' + this.x + '</b>'].concat(
                this.points ?
                    this.points.map(function (point) {
                        return point.series.name + ': ' + point.y + 'm';
                    }) : []
            );
        }"""),
        'split': True
     }, None),
    ({
        'formatter': """function () {
            // The first returned item is the header, subsequent items are the
            // points
            return ['<b>' + this.x + '</b>'].concat(
                this.points ?
                    this.points.map(function (point) {
                        return point.series.name + ': ' + point.y + 'm';
                    }) : []
            );
        }""",
        'split': True
     }, None),

    ({
      'border_width': 'not-a-number'
    }, (ValueError, TypeError)),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('tooltips/01.js', False, None),
    ('tooltips/02.js', False, None),

    ('tooltips/error-01.js', False, (TypeError, ValueError)),
    ('tooltips/error-02.js', False, errors.HighchartsParseError),

    ('tooltips/01.js', True, None),
    ('tooltips/02.js', True, None),

    ('tooltips/error-01.js', True, (TypeError, ValueError)),
    ('tooltips/error-02.js', True, errors.HighchartsParseError),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)
